
import numpy as np
from sklearn.preprocessing import StandardScaler
from sklearn.model_selection import train_test_split
from xgboost import XGBRegressor
from pymoo.algorithms.moo.nsga2 import NSGA2
from pymoo.core.problem import Problem
from pymoo.optimize import minimize


def load_data():
    """Load preprocessed data"""
    parameter_name = np.load('data/parameter_name.npy', allow_pickle=True)
    parameters = np.load('data/parameters.npy')
    data_hotwell = np.load('data/data_hotwell.npy')
    data_coldwell = np.load('data/data_coldwell.npy')
    data_caprock = np.load('data/data_caprock.npy')
    return parameter_name, parameters, data_hotwell, data_coldwell, data_caprock


def calculate_effective_heat_recovery(b_data, t, T_ref):
    """Calculate annual effective heat recovery rate."""
    samples, total_points = b_data.shape
    points_per_year = 40
    years = total_points // points_per_year
    heat_recovery_rates = np.zeros((samples, years))

    for i in range(samples):
        for year in range(years):
            start = year * points_per_year
            inj_start, inj_end = start + 20, start + 30
            prod_start, prod_end = start + 30, start + 40

            Q_inj = (t[i] - T_ref) * (inj_end - inj_start)
            Q_rec = np.sum(b_data[i, prod_start:prod_end] - T_ref)
            heat_recovery_rates[i, year] = Q_rec / Q_inj if Q_inj > 0 else 0
    return heat_recovery_rates

def calculate_annual_heat_production(b_data, mass_flow_rate):
    """Calculate annual heat production in GJ."""
    Cp = 4200  # J/kg°C
    T_ref = 30
    sec_per_year = 365 * 24 * 3600
    points_per_year = 40
    delta_t = sec_per_year / points_per_year

    samples, total_points = b_data.shape
    years = total_points // points_per_year
    heat_production = np.zeros((samples, years))

    for i in range(samples):
        for year in range(years):
            start = year * points_per_year + 30
            end = start + 10
            delta_T = b_data[i, start:end] - T_ref
            Q = np.sum(delta_T * delta_t) * mass_flow_rate[i] * Cp
            heat_production[i, year] = Q / 3.6e9
    return heat_production


# === Load and prepare data ===
parameter_name, parameters, data_hotwell, data_coldwell, data_caprock = load_data()

heat_recovery_rates = calculate_effective_heat_recovery(data_hotwell, t=parameters[:, 1], T_ref=30)
heat_production = calculate_annual_heat_production(data_hotwell, mass_flow_rate=parameters[:, 0])

c1 = heat_recovery_rates[:, 19:20]  # Heat recovery efficiency in year 20
c2 = heat_production[:, 19:20]      # Annual heat production (GJ) in year 20
c3 = data_coldwell[:, 79:80]        # Slip tendency (phi) at the cold well after cold water injection
c4 = data_caprock[:, 39:40]         # Slip tendency (phi) in the caprock after hot water injection

scaler_y1, scaler_y2, scaler_y3, scaler_y4 = StandardScaler(), StandardScaler(), StandardScaler(), StandardScaler()
c1_scaled = scaler_y1.fit_transform(c1)
c2_scaled = scaler_y2.fit_transform(c2)
c3_scaled = scaler_y3.fit_transform(c3)
c4_scaled = scaler_y4.fit_transform(c4)

# Sensitive parameter indices
X1 = parameters[:, [0, 11, 3, 13, 7]]
X2 = parameters[:, [0, 1, 7]]
X3 = parameters[:, [2, 7, 6, 18, 3]]
X4 = parameters[:, [1, 7, 18, 16, 6,5]]

scaler_x1, scaler_x2, scaler_x3, scaler_x4 = StandardScaler(), StandardScaler(), StandardScaler(), StandardScaler()
X1_scaled = scaler_x1.fit_transform(X1)
X2_scaled = scaler_x2.fit_transform(X2)
X3_scaled = scaler_x3.fit_transform(X3)
X4_scaled = scaler_x4.fit_transform(X4)

X_train_c1, X_test_c1, y_train_c1, y_test_c1 = train_test_split(X1_scaled, c1_scaled, test_size=0.2, random_state=42)
X_train_c2, X_test_c2, y_train_c2, y_test_c2 = train_test_split(X2_scaled, c2_scaled, test_size=0.2, random_state=42)
X_train_c3, X_test_c3, y_train_c3, y_test_c3 = train_test_split(X3_scaled, c3_scaled, test_size=0.2, random_state=42)
X_train_c4, X_test_c4, y_train_c4, y_test_c4 = train_test_split(X4_scaled, c4_scaled, test_size=0.2, random_state=42)

model_c1 = XGBRegressor(n_estimators=1000, max_depth=8, learning_rate=0.05,
                        subsample=0.8, colsample_bytree=0.8,
                        reg_alpha=10, reg_lambda=10, random_state=42)
model_c2 = XGBRegressor(n_estimators=1000, max_depth=8, learning_rate=0.05,
                        subsample=0.8, colsample_bytree=0.8,
                        reg_alpha=10, reg_lambda=10, random_state=42)
model_c3 = XGBRegressor(n_estimators=1000, max_depth=8, learning_rate=0.05,
                        subsample=0.8, colsample_bytree=0.8,
                        reg_alpha=10, reg_lambda=10, random_state=42)
model_c4 = XGBRegressor(n_estimators=1000, max_depth=8, learning_rate=0.05,
                        subsample=0.8, colsample_bytree=0.8,
                        reg_alpha=10, reg_lambda=10, random_state=42)

model_c1.fit(X_train_c1, y_train_c1.ravel())
model_c2.fit(X_train_c2, y_train_c2.ravel())
model_c3.fit(X_train_c3, y_train_c3.ravel())
model_c4.fit(X_train_c4, y_train_c4.ravel())

print("Model C1 R^2:", model_c1.score(X_test_c1, y_test_c1))
print("Model C2 R^2:", model_c2.score(X_test_c2, y_test_c2))
print("Model C3 R^2:", model_c3.score(X_test_c3, y_test_c3))
print("Model C4 R^2:", model_c4.score(X_test_c4, y_test_c4))

class MultiObjectiveOptimization(Problem):
    def __init__(self, y_fixed, model_c1, model_c2, model_c3, model_c4):
        super().__init__(n_var=3, n_obj=4, n_constr=0,
                         xl=np.array([5, 40, 10]),
                         xu=np.array([15, 180, 30]))
        self.y_fixed = y_fixed
        self.model_c1 = model_c1
        self.model_c2 = model_c2
        self.model_c3 = model_c3
        self.model_c4 = model_c4

    def _evaluate(self, x, out, *args, **kwargs):
        X_c1 = np.hstack([x[:, [0]], np.tile(self.y_fixed[[11, 3, 13, 7]], (x.shape[0], 1))])
        X_c2 = np.hstack([x[:, [0, 1]], np.tile(self.y_fixed[[7]], (x.shape[0], 1))])
        X_c3 = np.hstack([x[:, [2]], np.tile(self.y_fixed[[7, 6, 18, 3]], (x.shape[0], 1))])
        X_c4 = np.hstack([x[:, [1]], np.tile(self.y_fixed[[7, 18, 16, 6, 5]], (x.shape[0], 1))])

        f1 = -self.model_c1.predict(scaler_x1.transform(X_c1))
        f2 = -self.model_c2.predict(scaler_x2.transform(X_c2))
        f3 = self.model_c3.predict(scaler_x3.transform(X_c3))
        f4 = self.model_c4.predict(scaler_x4.transform(X_c4))
        out["F"] = np.column_stack([f1, f2, f3, f4])

y_fixed = np.array([10, 110, 20, 3e-13, 1e-17, 15, 15, 5e-6, 5e-7, 0.2,
                    0.015, 2.5, 3, 1150, 1500, 2150, 2500, 150, 800])

algorithm = NSGA2(pop_size=50)
problem = MultiObjectiveOptimization(y_fixed, model_c1, model_c2, model_c3, model_c4)

result = minimize(problem, algorithm, ('n_gen', 1000), seed=42, verbose=True)
pareto_solutions = result.X
pareto_objectives = result.F
